/*
 * Copyright 2012 Steven Gribble.
 *    [with contributions from cse333 students.]
 *
 *  This file is part of the UW CSE333 project sequence (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <unistd.h>
#include <errno.h>
#include <sys/select.h>

extern "C" {
  #include "./LinkedList.h"
  #include "./LinkedList_priv.h"
}

#include "./test_suite.h"
#include "./test_linkedlist.h"

namespace hw1 {

static unsigned int free_count = 0U;
void PayloadFreeFunction(void *payload) {
  // Do nothing but verify the payload is non-NULL and
  // increment the free count.
  ASSERT_NE(payload, (void *) NULL);
  free_count++;
}

int TestPayloadComparator(void *p1, void *p2) {
  // A comparator used to test sort.
  uint64_t i1 = (uint64_t) p1;
  uint64_t i2 = (uint64_t) p2;

  if (i1 > i2)
    return 1;
  if (i1 < i2)
    return -1;
  return 0;
}

TEST_F(Test_LinkedList, TestLinkedListBasic) {
  // Try creating a list.
  LinkedListPtr llp = AllocateLinkedList();
  ASSERT_NE((LinkedListPtr) NULL, llp);
  ASSERT_EQ(0U, NumElementsInLinkedList(llp));
  ASSERT_EQ(NULL, llp->head);
  ASSERT_EQ(NULL, llp->tail);
  HW1Addpoints(5);

  // Try deleting the (empty) list.
  free_count = 0U;
  FreeLinkedList(llp, &PayloadFreeFunction);
  ASSERT_EQ(0U, free_count);
  llp = NULL;
  HW1Addpoints(5);
}

TEST_F(Test_LinkedList, TestLinkedListPushPop) {
  // Creating a list.
  LinkedListPtr llp = AllocateLinkedList();
  ASSERT_NE((LinkedListPtr) NULL, llp);
  ASSERT_EQ(0U, NumElementsInLinkedList(llp));
  ASSERT_EQ(NULL, llp->head);
  ASSERT_EQ(NULL, llp->tail);

  // Insert an element.
  ASSERT_EQ(1, PushLinkedList(llp, (void *) 1));
  ASSERT_EQ(1U, NumElementsInLinkedList(llp));
  ASSERT_EQ(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ((void *) 1, llp->head->payload);
  HW1Addpoints(10);

  // Delete the element.
  void *payload_ptr;
  ASSERT_EQ(1, PopLinkedList(llp, &payload_ptr));
  ASSERT_EQ((void *) 1, payload_ptr);
  ASSERT_EQ(0U, NumElementsInLinkedList(llp));
  HW1Addpoints(10);

  // Delete the element a second time.
  ASSERT_EQ(0, PopLinkedList(llp, &payload_ptr));
  HW1Addpoints(5);

  // Insert two elements.
  ASSERT_EQ(1, PushLinkedList(llp, (void *) 1));
  ASSERT_EQ(1U, NumElementsInLinkedList(llp));
  ASSERT_EQ(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ((void *) 1, llp->head->payload);

  ASSERT_EQ(1, PushLinkedList(llp, (void *) 2));
  ASSERT_EQ(2U, NumElementsInLinkedList(llp));
  ASSERT_NE(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ(llp->tail, llp->head->next);
  ASSERT_EQ(llp->head, llp->tail->prev);
  ASSERT_EQ((void *) 2, llp->head->payload);
  ASSERT_EQ((void *) 1, llp->tail->payload);
  HW1Addpoints(10);

  // Delete the first element.
  ASSERT_EQ(1, PopLinkedList(llp, &payload_ptr));
  ASSERT_EQ((void *) 2, payload_ptr);
  ASSERT_EQ(1U, NumElementsInLinkedList(llp));
  ASSERT_EQ(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ((void *) 1, llp->head->payload);
  HW1Addpoints(10);

  // Delete the non-empty list.
  free_count = 0U;
  FreeLinkedList(llp, &PayloadFreeFunction);
  ASSERT_EQ(1U, free_count);
  llp = NULL;
}

TEST_F(Test_LinkedList, TestLinkedListAppendSlice) {
  // Creating a list.
  LinkedListPtr llp = AllocateLinkedList();
  ASSERT_NE((LinkedListPtr) NULL, llp);
  ASSERT_EQ(0U, NumElementsInLinkedList(llp));
  ASSERT_EQ(NULL, llp->head);
  ASSERT_EQ(NULL, llp->tail);

  // Insert an element.
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 1));
  ASSERT_EQ(1U, NumElementsInLinkedList(llp));
  ASSERT_EQ(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ((void *) 1, llp->head->payload);
  HW1Addpoints(5);

  // Delete the element.
  void *payload_ptr;
  ASSERT_EQ(1, SliceLinkedList(llp, &payload_ptr));
  ASSERT_EQ((void *) 1, payload_ptr);
  ASSERT_EQ(0U, NumElementsInLinkedList(llp));
  HW1Addpoints(5);

  // Delete the element a second time.
  ASSERT_EQ(0, SliceLinkedList(llp, &payload_ptr));
  HW1Addpoints(5);

  // Insert two elements.
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 1));
  ASSERT_EQ(1U, NumElementsInLinkedList(llp));
  ASSERT_EQ(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ((void *) 1, llp->head->payload);

  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 2));
  ASSERT_EQ(2U, NumElementsInLinkedList(llp));
  ASSERT_NE(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ(llp->tail, llp->head->next);
  ASSERT_EQ(llp->head, llp->tail->prev);
  ASSERT_EQ((void *) 1, llp->head->payload);
  ASSERT_EQ((void *) 2, llp->tail->payload);
  HW1Addpoints(5);

  // Delete the first element.
  ASSERT_EQ(1, SliceLinkedList(llp, &payload_ptr));
  ASSERT_EQ((void *) 2, payload_ptr);
  ASSERT_EQ(1U, NumElementsInLinkedList(llp));
  ASSERT_EQ(llp->head, llp->tail);
  ASSERT_EQ(NULL, llp->head->prev);
  ASSERT_EQ(NULL, llp->tail->next);
  ASSERT_EQ((void *) 1, llp->head->payload);
  HW1Addpoints(5);

  // Delete the non-empty list.
  free_count = 0U;
  FreeLinkedList(llp, &PayloadFreeFunction);
  ASSERT_EQ(1U, free_count);
  llp = NULL;
}

TEST_F(Test_LinkedList, TestLinkedListSort) {
  // Creating a list.
  LinkedListPtr llp = AllocateLinkedList();
  ASSERT_NE((LinkedListPtr) NULL, llp);
  ASSERT_EQ(0U, NumElementsInLinkedList(llp));
  ASSERT_EQ(NULL, llp->head);
  ASSERT_EQ(NULL, llp->tail);

  // Insert some elements.
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 3));
  ASSERT_EQ(1U, NumElementsInLinkedList(llp));
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 2));
  ASSERT_EQ(2U, NumElementsInLinkedList(llp));
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 1));
  ASSERT_EQ(3U, NumElementsInLinkedList(llp));

  // Sort ascending.
  SortLinkedList(llp, 1, &TestPayloadComparator);

  // Verify the sort.
  ASSERT_EQ(1UL, (uint64_t) llp->head->payload);
  ASSERT_EQ(2UL, (uint64_t) llp->head->next->payload);
  ASSERT_EQ(3UL, (uint64_t) llp->head->next->next->payload);
  ASSERT_EQ(NULL, llp->head->next->next->next);

  // Resort descending.
  SortLinkedList(llp, 0, &TestPayloadComparator);

  // Verify the sort.
  ASSERT_EQ(3UL, (uint64_t) llp->head->payload);
  ASSERT_EQ(2UL, (uint64_t) llp->head->next->payload);
  ASSERT_EQ(1UL, (uint64_t) llp->head->next->next->payload);
  ASSERT_EQ(NULL, llp->head->next->next->next);
  HW1Addpoints(5);

  // Delete the non-empty list.
  free_count = 0U;
  FreeLinkedList(llp, &PayloadFreeFunction);
  ASSERT_EQ(3U, free_count);
  llp = NULL;
}

TEST_F(Test_LinkedList, TestLLIteratorBasic) {
  // Create a linked list.
  LinkedListPtr llp = AllocateLinkedList();

  // Make sure you can't create an iterator to an empty list.
  ASSERT_EQ(NULL, LLMakeIterator(llp, 0));
  HW1Addpoints(5);

  // Add some data to the list.
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 3));
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 2));
  ASSERT_EQ(1, AppendLinkedList(llp, (void *) 1));

  // Create the iterator.
  LLIterPtr lli = LLMakeIterator(llp, 0);
  ASSERT_NE((LLIterPtr) NULL, lli);
  ASSERT_EQ(1, LLIteratorHasNext(lli));
  ASSERT_EQ(0, LLIteratorHasPrev(lli));
  ASSERT_EQ(llp, lli->list);
  ASSERT_EQ(llp->head, lli->node);
  HW1Addpoints(5);

  // Navigate using the iterator.
  void *payload;
  ASSERT_EQ(0, LLIteratorPrev(lli));
  ASSERT_EQ(1, LLIteratorNext(lli));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(2U, (uint64_t) payload);
  ASSERT_EQ(1, LLIteratorHasNext(lli));
  ASSERT_EQ(1, LLIteratorHasPrev(lli));
  ASSERT_EQ(1, LLIteratorNext(lli));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(1U, (uint64_t) payload);
  ASSERT_EQ(0, LLIteratorHasNext(lli));
  ASSERT_EQ(1, LLIteratorHasPrev(lli));
  ASSERT_EQ(0, LLIteratorNext(lli));
  ASSERT_EQ(1, LLIteratorPrev(lli));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(2U, (uint64_t) payload);
  HW1Addpoints(10);

  // Do an insert-before from the middle element.
  ASSERT_EQ(1, LLIteratorInsertBefore(lli, (void *) 4));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(2U, (uint64_t) payload);
  ASSERT_EQ(1, LLIteratorPrev(lli));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(4U, (uint64_t) payload);

  // Rewind to beginning, try insert-before.
  ASSERT_EQ(1, LLIteratorPrev(lli));
  ASSERT_EQ(0, LLIteratorHasPrev(lli));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(3U, (uint64_t) payload);
  ASSERT_EQ(1, LLIteratorInsertBefore(lli, (void *) 5));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(3U, (uint64_t) payload);
  ASSERT_EQ(1, LLIteratorPrev(lli));
  ASSERT_EQ(0, LLIteratorHasPrev(lli));
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(5U, (uint64_t) payload);
  HW1Addpoints(10);

  // The list contains 5 elements. Try a delete from the front of the list.
  // (i.e., delete element 1/5.)
  free_count = 0;
  LinkedListNodePtr prev, next = lli->node->next;
  ASSERT_EQ(1, LLIteratorDelete(lli, &PayloadFreeFunction));
  ASSERT_EQ(next, lli->node);
  ASSERT_EQ(NULL, lli->node->prev);
  ASSERT_EQ(4U, NumElementsInLinkedList(lli->list));
  ASSERT_EQ(1U, free_count);
  LLIteratorGetPayload(lli, &payload);
  ASSERT_EQ(3U, (uint64_t) payload);
  ASSERT_EQ(0, LLIteratorHasPrev(lli));
  HW1Addpoints(10);

  // Delete the rest.
  // Move the iterator forward by one to test removing from the middle.
  // Deleting element 2/4
  ASSERT_EQ(1, LLIteratorNext(lli));
  prev = lli->node->prev;
  next = lli->node->next;
  ASSERT_EQ(1, LLIteratorDelete(lli, &PayloadFreeFunction));
  ASSERT_EQ(3U, NumElementsInLinkedList(lli->list));
  ASSERT_EQ(next, lli->node);
  ASSERT_EQ(prev, lli->node->prev);

  // Delete from the middle. Deleting element 2/3
  next = lli->node->next;
  prev = lli->node->prev;
  ASSERT_EQ(1, LLIteratorDelete(lli, &PayloadFreeFunction));
  ASSERT_EQ(2U, NumElementsInLinkedList(lli->list));
  ASSERT_EQ(next, lli->node);
  ASSERT_EQ(prev, lli->node->prev);

  // Delete from the tail position.
  // Deleting element 2/2
  prev = lli->node->prev;
  ASSERT_EQ(1, LLIteratorDelete(lli, &PayloadFreeFunction));
  ASSERT_EQ(1U, NumElementsInLinkedList(lli->list));
  ASSERT_EQ(NULL, lli->node->next);
  ASSERT_EQ(prev, lli->node);

  // Remove the remaining node from the list. Deleting element 1/1.
  ASSERT_EQ(0, LLIteratorDelete(lli, &PayloadFreeFunction));
  ASSERT_EQ(0U, NumElementsInLinkedList(lli->list));
  ASSERT_EQ(NULL, lli->node);
  ASSERT_EQ(5U, free_count);

  // Free the iterator.
  LLIteratorFree(lli);
  HW1Addpoints(5);

  // Free the list.
  FreeLinkedList(llp, &PayloadFreeFunction);
}

}  // namespace hw1
